//Boston University - Electronics Design Facility
//Monkey Timer Mk2
//http://edf.bu.edu/
//written by Chris Lawlor, January - April 2011
//Device is Atmega2560
//CPU clock is 16Mhz
//This file defines the major measurement data structures and data access functions

#ifndef _MONKEY_MK2_DATATYPES_H
#define _MONKEY_MK2_DATATYPES_H 1 

#include <inttypes.h>
#include <avr/eeprom.h>

typedef uint8_t	result_t;


// Timestamp
typedef struct a_time { 
	uint8_t month; 	// month 1-12
	uint8_t mday; 	// day 1-31
	uint8_t year; 	// year - 2000
	uint8_t hour; 	// hour 0-23
	uint8_t min; 	// minute 0-59
	uint8_t sec; 	// seconds 0-59
} a_time_t;


// One timer measurement
typedef struct a_measure { 
	a_time_t time;
	uint16_t interval; 	// interval in 1/100 sec 0-5999
						// it would be good to use integers 0000 to 5999 for 00.00 to 59.99 
	uint8_t flags; 		// bit field with flags
						// Left/Right -> bit0 1/0, Valid/Invalid -> bit1 1/0
						// Downloaded Flag
						// high byte is animal number
} a_measure_t;


#define LEFT_MASK		0x01	//Left/Right -> bit0 1/0
#define VALID_MASK		0x02	//Valid/Invalid -> bit1 1/0
#define IS_DOWNLOADED	0x04	//this flag will be set after the data has been downloaded
#define ANIMAL_NUM_MSK	0xF8	//the animal number 1 thru 32
#define INTERVAL_SCALE	100		//LSB of interval is (1/INTERVAL_SCALE) seconds (should be a power of 10)


#define MAX_ANIMALS 31
uint8_t	 EEMEM AnimalNumber = 0; //0 - 31


uint8_t get_animal_number()
{
	uint8_t animal_num = 255;
	uint8_t sreg_save = SREG;
	cli();
	animal_num = eeprom_read_byte(&AnimalNumber);
	SREG = sreg_save;
	return animal_num;
}

void set_animal_number(uint8_t animal_num)
{
	uint8_t sreg_save = SREG;
	cli();
	eeprom_write_byte(&AnimalNumber, animal_num);
	SREG = sreg_save;
	return;

}

//a_measure_t testvect1 = {{1, 2, 11, 11, 12, 13},599,0x02};
//a_measure_t testvect2 = {{1, 2, 11, 11, 12, 13},233,0x01};


/*
uint8_t	 EEMEM testchar = 0x54;		//'T'
uint16_t EEMEM testword = 0x4948;	//"HI"
uint8_t  EEMEM teststring[10] = {'q','w','e','r','t','y','u','i','o','p'};
a_measure_t EEMEM t1;
*/

void write_a_measure_t (a_measure_t *source_ram, a_measure_t *dest_eep)
{
	uint8_t copy_SREG = SREG;
	cli();
	eeprom_write_block ( source_ram, dest_eep, sizeof(a_measure_t));
	SREG = copy_SREG;
}

void read_a_measure_t (a_measure_t *dest_ram, a_measure_t *source_eep)
{	
	uint8_t copy_SREG = SREG;
	cli();
	eeprom_read_block ( dest_ram, source_eep, sizeof(a_measure_t)); 
	SREG = copy_SREG;
}

int a_time_t_ToString(a_time_t *t, char *s, int N)
//returns the length of the string not including the null term char
//returns < 0 on error or overflow
{
	return snprintf(s, N, "%02d/%02d/%04d  %02d:%02d:%02d", t->month, t->mday, (t->year+2000), t->hour, t->min, t->sec);	
}

int a_measure_t_ToString(a_measure_t *a, char *s, int N)
//returns the length of the string not including the null term char
//returns < 0 on error or overflow
{
	//int p = a_time_t_ToString(&a->time, s, N);
	//if (p<0)
	//	return p;

	snprintf(s, N, "%02d/%02d/%04d,%02d:%02d:%02d", a->time.month, a->time.mday, (a->time.year+2000), a->time.hour, a->time.min, a->time.sec);
	return snprintf(s, N, "%s,%u,%d.%02d,%c,%c,%c\n", s, ((a->flags&ANIMAL_NUM_MSK)>>3),(a->interval/INTERVAL_SCALE), (a->interval%INTERVAL_SCALE), ((a->flags&LEFT_MASK)?'L':'R'), ((a->flags&VALID_MASK)?'Y':'N'), ((a->flags&IS_DOWNLOADED)?'Y':'N'));
}

int a_measure_t_ToDisplayString(a_measure_t *a, char *s, int N)
//returns the length of the string not including the null term char
//returns < 0 on error or overflow
{
	return snprintf(s, N, " %02d:%02d:%02d %02d.%02d %c %c", a->time.hour, a->time.min, a->time.sec, (a->interval/INTERVAL_SCALE), (a->interval%INTERVAL_SCALE), ((a->flags&LEFT_MASK)?'L':'R'), ((a->flags&VALID_MASK)?'Y':'N'));
}


void a_time_t_Incriment(a_time_t *t) 
{
	t->sec++; 	// seconds 0-59
	if (t->sec == 60)
	{	
		t->sec = 0;
		t->min++; //minutes 0-59
		if (t->min == 60)
		{	
			t->min = 0;
			t->hour++; //hours 0-23
			if (t->hour == 24)
			{	
				t->hour = 0;
				t->mday++; 	//day 1-31 depending on the month
				if ((t->month == 9)||(t->month == 4)||(t->month == 6)||(t->month == 11))
				{
					if (t->mday==31)
					{
						t->mday=1;
						t->month++;
					}
				}
				else if (t->month == 2)
				{
					if ( (t->year%4) == 0)
					{
						//leap year
						if (t->mday==30)
						{
							t->mday=1;
							t->month++;
						}
					}
					else 
					{
						if (t->mday==29)
						{
							t->mday=1;
							t->month++;
						}
					}
				}
				else
				{
					if (t->mday==32)
					{
						t->mday=1;
						t->month++; //month 1-12
						if (t->month == 13)
						{
							t->month = 1;
							t->year++;
						}
					}
				}
			}
		}
	}
}



uint8_t is_valid_date(a_time_t *t)
//checks the date for validity
{
	// seconds 0-59
	if ((t->sec >= 0)&&(t->sec < 60))
	{	
		 //minutes 0-59
		if ((t->min >= 0)&&(t->min < 60))
		{	
			//hours 0-23
			if ((t->hour >= 0)&&(t->hour < 24))
			{	
				//day 1-31 depending on the month
				if ((t->month == 9)||(t->month == 4)||(t->month == 6)||(t->month == 11))
				{
					if ((t->mday > 0)&&(t->mday < 31))
					{
						return 1;
					}
				}
				else if (t->month == 2)
				{
					if ( (t->year%4) == 0)
					{
						//leap year
						if ((t->mday > 0)&&(t->mday < 30))
						{
							return 1;
						}
					}
					else 
					{
						if ((t->mday > 0)&&(t->mday < 29))
						{
							return 1;
						}
					}
				}
				else if ((t->month == 1)||(t->month == 3)||(t->month == 5)||(t->month == 7)||(t->month == 8)||(t->month == 10)||(t->month == 12))
				{
					if ((t->mday > 0)&&(t->mday < 32))
					{
						return 1;
					}
				}
			}
		}
	}

	return 0;
}


//--------------------------------------------------------------------------
//--------------------------------------------------------------------------


#define NUM_MEASUREMENTS		400
#define EEP_FLAGS_FIFO_FULL		0x01		
#define EEP_FLAGS_FIFO_EMPTY	0x02

a_measure_t EEMEM eep_fifo[NUM_MEASUREMENTS];

uint16_t 	EEMEM  	eep_fifo_first = 0;
uint16_t	EEMEM  	eep_fifo_last = 0;
uint8_t		EEMEM	eep_fifo_flags = EEP_FLAGS_FIFO_EMPTY;


uint16_t circular_inc(uint16_t *i)
{
	(*i)++;
	if (*i==NUM_MEASUREMENTS)
		*i = 0;
	return *i;
}

uint16_t circular_dec(uint16_t *i)
{
	(*i)--;
	if (*i<0)
		*i = NUM_MEASUREMENTS - 1;
	return *i; 
}


int8_t eep_fifo_push_last(a_measure_t *data)
//adds *data to the end of the Fifo
//returns -1 on error.
{
	uint8_t copy_SREG = SREG;
	cli();
	uint8_t flags = eeprom_read_byte(&eep_fifo_flags);
	if (flags&EEP_FLAGS_FIFO_FULL)
	{
		SREG = copy_SREG;
		return -1;	
	}
	uint16_t last = eeprom_read_word(&eep_fifo_last);
	uint16_t first = eeprom_read_word(&eep_fifo_first);
	write_a_measure_t(data, &eep_fifo[last]);
	circular_inc(&last);
	eeprom_write_word(&eep_fifo_last, last);

	if 	(flags&EEP_FLAGS_FIFO_EMPTY)
	{
		flags &= ~EEP_FLAGS_FIFO_EMPTY;
	}
	
	if (last==first)
	{
		flags |= EEP_FLAGS_FIFO_FULL;
	}

	eeprom_write_byte(&eep_fifo_flags, flags);
	SREG = copy_SREG;
	return 1;
}

int8_t eep_fifo_pop_first(a_measure_t *data)
//removes the first element of the Fifo and returns it in the 
//struct pointed to by *data.  
//data must point to a valid variable of a_measure_t type.
//returns -1 on error.
{
	uint8_t copy_SREG = SREG;
	cli();
	uint8_t flags = eeprom_read_byte(&eep_fifo_flags);
	if (flags&EEP_FLAGS_FIFO_EMPTY)
	{
		SREG = copy_SREG;
		return -1;	
	}
	uint16_t last = eeprom_read_word(&eep_fifo_last);
	uint16_t first = eeprom_read_word(&eep_fifo_first);
	read_a_measure_t(data, &eep_fifo[first]);
	circular_inc(&first);
	eeprom_write_word(&eep_fifo_first, first);

	if (last==first)
	{
		flags |= EEP_FLAGS_FIFO_EMPTY;
	}

	flags &= ~EEP_FLAGS_FIFO_FULL;

	eeprom_write_byte(&eep_fifo_flags, flags);
	SREG = copy_SREG;
	return 1;
}

void eep_fifo_clear()
//a call to eep_fifo_clear() clears the entire fifo
{
	uint8_t copy_SREG = SREG;
	cli();
	eeprom_write_word(&eep_fifo_first, 0);
	eeprom_write_word(&eep_fifo_last, 0);
	eeprom_write_byte(&eep_fifo_flags, EEP_FLAGS_FIFO_EMPTY);
	SREG = copy_SREG;
}

uint8_t eep_fifo_is_full()
{
	uint8_t copy_SREG = SREG;
	cli();
	uint8_t flags = eeprom_read_byte(&eep_fifo_flags);
	SREG = copy_SREG;
	return (flags&EEP_FLAGS_FIFO_FULL);
}

uint8_t eep_fifo_is_empty()
{
	uint8_t copy_SREG = SREG;
	cli();
	uint8_t flags = eeprom_read_byte(&eep_fifo_flags);
	SREG = copy_SREG;
	return (flags&EEP_FLAGS_FIFO_EMPTY);
}

int8_t eep_fifo_peek_last(a_measure_t *data, uint16_t n)
//returns the n-th from the last element
//n==0 will get the last element
//n==1 will get the next to last, etc...
//returns 
//-1 on fifo empty
//-2 on no more data (ie, n too big)
{
	uint8_t copy_SREG = SREG;
	cli();
	uint8_t flags = eeprom_read_byte(&eep_fifo_flags);

	if (flags&EEP_FLAGS_FIFO_EMPTY)
	{
		SREG = copy_SREG;
		return -1;	
	}

	uint16_t last = eeprom_read_word(&eep_fifo_last);
	uint16_t first = eeprom_read_word(&eep_fifo_first);
	uint16_t i = last;
	circular_dec(&i);

	while (n>0 && i!=first)
	{
		circular_dec(&i);
		n--;
	}

	if (n>0)
	{
		SREG = copy_SREG;
		return -2;	
	}

	read_a_measure_t(data, &eep_fifo[i]);
	
	SREG = copy_SREG;
	return 1;
}

#endif
