//Boston University - Electronics Design Facility
//Monkey Timer Mk2
//http://edf.bu.edu/
//written by Chris Lawlor, January - April 2011
//Device is Atmega2560
//CPU clock is 16Mhz
//This file defines the behaviour of the frontpanel buttons for the Monkey Timer Mk2

#ifndef _MONKEY_MK2_FRONTPANEL_BUTTONS_H
#define _MONKEY_MK2_FRONTPANEL_BUTTONS_H 1


//There are 4 buttons on the Monkey Timer Mk2:  UP, DOWN, LEFT_MENU_RIGHT, and SELECT
//
//1. UP is a momentary pushbutton -- 
//		it moves the menu selection cursor up, or incriments a selected value
//
//2. DOWN is a momentary pushbutton -- 
//		it moves the menu selection cursor down, or decriments a selected value
//
//3. LEFT_MENU_RIGHT is a 3-way toggle switch --
//		the Left position selects left channel mode, 
//		the Middle position selects menu mode,
//		the Right position selects right channel mode
//
//4. SELECT is a momentary pushbutton --
// 		in L or R channel mode then it's a START button, in Menu mode then it's a SELECT button


#define PORTF_BUTTON_MASK	0x1F
#define BUTTONS				(PINF & PORTF_BUTTON_MASK)

#define FP_BUTTON_UP_MSK			0x10
#define FP_BUTTON_DN_MSK			0x08
#define FP_BUTTON_START_SEL_MSK		0x04
#define FP_BUTTON_LEFT_MSK			0x02
#define FP_BUTTON_RIGHT_MSK			0x01
#define FP_BUTTON_LRM_MSK			(FP_BUTTON_RIGHT_MSK|FP_BUTTON_LEFT_MSK)
#define	FP_BUTTON_LOCKOUT_PERIOD	10		//for software switch debouncing, (a person can only physically push a button at ~5Hz)


volatile uint8_t	fp_buttons;
volatile uint8_t 	new_buttons;
volatile uint16_t	fp_button_update_lockout;
volatile uint8_t	fp_button_state_isnew;

#define BUTTON_HANDLER_PERIOD	156	//156 corresponds to ~10ms, if button presses are being missed, make it less

void frontpanel_buttons_init()
{
	DDRF &= ~PORTF_BUTTON_MASK;	//set them all as inputs
	fp_buttons = 0xFF;
	fp_button_update_lockout = 0;
	fp_button_state_isnew = 0;

	//setup button handler interrupt
	TCCR2A = 0x02;
	OCR2A  = BUTTON_HANDLER_PERIOD;
	TIMSK2 = 0x02;
	TCCR2B = 0x07;

	//TCCR4A = 0x00;
	//TCCR4C = 0x00;
	//OCR4A  = BUTTON_HANDLER_PERIOD;
	//TIMSK4 = 0x02;
	//TCCR4B = 0x0D;
}


inline void update_button_state()
//only call this from the ISR with interrupts blocked to avoid nested calls
{
	if (fp_button_update_lockout > 0) 
	{
		fp_button_update_lockout--;
	} 
	else 
	{	
		fp_button_update_lockout = FP_BUTTON_LOCKOUT_PERIOD;
		if (fp_button_state_isnew == 0)
		{
			new_buttons = BUTTONS;
			if ( (new_buttons != fp_buttons) )
			{
				fp_button_state_isnew = 1;
				fp_buttons = new_buttons;
			}
		}
	}
}


ISR( TIMER2_COMPA_vect, ISR_NOBLOCK)
{
	//DDRB |= 0x80;
	//PINB |= 0x80;
	update_button_state();
}



uint8_t	get_button_state()
{
	uint8_t copy_SREG = SREG;
	cli();
	volatile uint8_t temp = fp_buttons;
	fp_button_state_isnew = 0;
	SREG = copy_SREG;
	return temp;
}



//Button Access Macros
#define UP(t)			(!((t)&FP_BUTTON_UP_MSK))
#define DOWN(t)			(!((t)&FP_BUTTON_DN_MSK))
#define START(t)		(!((t)&FP_BUTTON_START_SEL_MSK))
#define SELECT(t)		START(t)
#define LEFT(t)			(!((t)&FP_BUTTON_LEFT_MSK))
#define RIGHT(t)		(!((t)&FP_BUTTON_RIGHT_MSK))
#define MENUBTN(t)		((!(LEFT(t)))&&(!(RIGHT(t))))
#define NEWBUTTON		(fp_button_state_isnew)


//these ones are more usefull
#define VALID_UP(b) 			( UP(b) && !DOWN(b) && !SELECT(b) )	//a primary keying
#define VALID_DOWN(b)			(!UP(b) &&  DOWN(b) && !SELECT(b) )	//a primary keying
#define VALID_SELECT(b) 		(!UP(b) && !DOWN(b) &&  SELECT(b) )	//a primary keying
#define IS_PUSHBUTTON_RETURN(b)	(!UP(b) && !DOWN(b) && !SELECT(b) )	
#define UP_DOWN_SELECT(b)		( UP(b) &&  DOWN(b) &&  SELECT(b) )	
#define UP_DOWN(b)				( UP(b) &&  DOWN(b) && !SELECT(b) )	
#define UP_SELECT(b)			( UP(b) && !DOWN(b) &&  SELECT(b) )	
#define DOWN_SELECT(b)			(!UP(b) &&  DOWN(b) &&  SELECT(b) )	



uint8_t get_valid_pushbutton()
//blocks until newbuttons reads them 
//loop until its a valid pushbutton
{
	uint8_t b;
	do {
		while(!NEWBUTTON)
		{}
		b = get_button_state();
	} while (!(VALID_UP(b)||VALID_DOWN(b)||VALID_SELECT(b)));
	return b;
}


//-----------------------------------------------------------------





void testbuttons()
{
	uint8_t	b;

	while(1)
	{	
		if (NEWBUTTON)
		{
			b = get_button_state();

			if (UP(b))
			{
				USART_SendString("UP\n",0);
			}			
			
			if (DOWN(b))
			{
				USART_SendString("DOWN\n",0);
			}
			
			if (LEFT(b))
			{
				USART_SendString("LEFT\n",0);
			}	
					
			if (RIGHT(b))
			{
				USART_SendString("RIGHT\n",0);
			}
				
			if (MENUBTN(b))
			{
				USART_SendString("MENUBTN\n",0);
	
			}	

			if (START(b))
			{
				USART_SendString("START\n",0);
			}


			if (SELECT(b))
			{
				USART_SendString("SELECT\n",0);
			}

		}
	}
}





#endif
