//Boston University - Electronics Design Facility
//Monkey Timer Mk2
//http://edf.bu.edu/
//written by Chris Lawlor, January - April 2011
//Device is Atmega2560
//CPU clock is 16Mhz
//This file defines interface to the TWI (aka I2C) for the ATMEG2560


#ifndef _TWI_ATMEGA2560_H
#define _TWI_ATMEGA2560_H 1



//function prototypes
void twi_master_setup();
//to set baud at 100kHz

int twi_master_read (char slave_address, char starting_address, char *buffer, int N);
//reads at most N bytes from the device located at slave_address and stores them in buffer
//the return value is the number of bytes actually read, or negative indicating error


int twi_master_write(char slave_address, char starting_address, char *buffer, int N);
//writes at most N bytes to the device located at slave_address from buffer
//the return value is the number of bytes actually read, or negative indicating error


#define TWI_READ	1
#define TWI_WRITE	0

#define TWI_STAUS_MASK	0xF8
#define TWI_START_SENT	0x08
#define TWI_RPSRT_SENT	0x10
#define	TWI_SLAR_ACK	0x40
#define	TWI_SLAR_NACK	0x48
#define	TWI_SLAW_ACK	0x18
#define	TWI_SLAW_NACK	0x20
#define	TWI_DATAR_ACK	0x50
#define	TWI_DATAR_NACK	0x58
#define	TWI_DATAS_ACK	0x28
#define	TWI_DATAS_NACK	0x30
#define	TWI_ARB_LOST	0x38


void twi_master_setup()
//to set baud at 100kHz
{
	TWBR = 18;
	TWCR = 0x44;
	TWSR = 0x01;	//this line works, but seems contrary to the datasheet
}

int twi_master_read (char slave_address, char starting_address, char *in_buffer, int N)
//reads N bytes from i2c device at slave_address beginning with offset starting_address and stores them in in_buffer
{	
	if (N==0)
		return 0;

	int i = 0;

	//1. issue the START condition
	TWCR |= (1<<TWSTA);
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_START_SENT);


	//2. address the slave device
	TWDR = ((slave_address<<1)|TWI_WRITE);
	TWCR = ((TWCR&~(1<<TWSTA))|(1<<TWINT));
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_SLAW_ACK);	

	//3. write the starting_address 
	TWDR = starting_address;
	TWCR |= (1<<TWINT);
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_DATAS_ACK);


	//4. NOW ISSUE REPEATED START 
	TWCR |= (1<<TWSTA)|(1<<TWINT);
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_RPSRT_SENT);


	//5. And RE-ADDRESS the slave
	TWDR = ((slave_address<<1)|TWI_READ);
	TWCR = ((TWCR&~(1<<TWSTA))|(1<<TWINT));	
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_SLAR_ACK);

	//6. now read the N bytes into the buffer
	while (i<N) //&& no errors)
	{
		if (i<(N-1))
		{
			TWCR |= (1<<TWINT);
			do {
				while (!(TWCR&(1<<TWINT)))
				{}
			}
			while ((TWSR&TWI_STAUS_MASK)!=TWI_DATAR_ACK);
		}
		else
		{
			//to mark the end of the multibyte read, you must NACK the final byte to singal the end of transmission to the slave transmitter
			//to send a NACK you must toggle the TWEA bit (THIS IS UNDOCUMENTED IN THE ATMEGA2560 MANUAL)
			//set it low here and then reset it while sending the stop condition 
			TWCR = (TWCR&~(1<<TWEA))|(1<<TWINT);
			do {
				while (!(TWCR&(1<<TWINT)))
				{}
			}
			while ((TWSR&TWI_STAUS_MASK)!=TWI_DATAR_NACK);
		}

		in_buffer[i] = TWDR;	
		i++;
	}

	//7. issue a stop condition
	TWCR |= (1<<TWEA)|(1<<TWINT)|(1<<TWSTO);
	
	return i;
}


int twi_master_write(char slave_address, char starting_address, char *out_buffer, int N)
//writes N bytes from out_buffer to the i2c device at slave_address with the slave's offset starting_address 
{
	if (N==0)
		return 0;

	int i = 0;
	//int NACK_RECEIVED = 0;

	//1. issue the START condition
	TWCR |= (1<<TWSTA)|(1<<TWINT);
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_START_SENT);

	//2. address the slave device
	TWDR = ((slave_address<<1)|TWI_WRITE);
	TWCR = ((TWCR&~(1<<TWSTA))|(1<<TWINT));
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_SLAW_ACK);

	//3. write the starting_address 
	TWDR = starting_address;
	TWCR |= (1<<TWINT);
	do {
		while (!(TWCR&(1<<TWINT)))
		{}
	}
	while ((TWSR&TWI_STAUS_MASK)!=TWI_DATAS_ACK);

	//4. now write the N bytes into the buffer
	while (i<N) //&& no errors)
	{
		TWDR = out_buffer[i];
		i++;
		TWCR |= (1<<TWINT);
		do {
			if (((TWSR&TWI_STAUS_MASK)==TWI_DATAS_NACK))
				break;
			while (!(TWCR&(1<<TWINT)))
			{}
		}
		while ((TWSR&TWI_STAUS_MASK)!=TWI_DATAS_ACK);
	}

	//5. issue a stop condition
	TWCR |= (1<<TWINT)|(1<<TWSTO);
	return i;
}


/*
void twi_toggle_pins()
//toggles the TWI pins at 100kHz for debuging
{
	DDRD = 0x03;
	//TCCR0A = 0x02;
	//OCR0A = 10;
	//TCCR0B = 0x02;
	while (1)
	{
		//while (!(TIFR0&(1<<OCF0A)))
		//{}
		//TIFR0 |= (1<<OCF0A);
		PIND = 0x03;
	}
}
*/


#endif
